module Typecheck (typeCheckExp, typeCheckStm, typeCheckVar) where 


import Abssyntax
import Control.Monad.State
import Data.Map as Map
import Prelude hiding (lookup)

type Types = [Map Ident Type]

inList :: Eq a => a -> [a] -> Bool
inList _ [] = False
inList a (x:xs) = if a == x then True else inList a xs

typeCheckExp :: (MonadState Types m) => Exp -> m Type
typeCheckExp (BiOpExp e o e') = do
	t1 <- typeCheckExp e
	t2 <- typeCheckExp e'
	if not(t1 == t2) then fail "The parameters for the binary operator aren't equal" 
		else if inList o [Eq,NEq] then return TBool
			else if not(t1 == TInt) then fail "The parameters need to be of type int" 
				else if inList o [Plus,Minus,Times,Div]
					then return TInt
					else return TBool
typeCheckExp (EVar i) = typeCheckVar i
typeCheckExp (EAss i e) = do
	t <- typeCheckVar i
	t2 <- typeCheckExp e
	if t == t2 then return t else fail $ "Illegal to assign an expression of type "++show t2++" to variable "++show i++" of type "++show t 
typeCheckExp (EInt i) = return TInt
typeCheckExp (EBool b) = return TBool
typeCheckExp EReadI = return TInt
typeCheckExp EReadB = return TBool
typeCheckExp (ExpT t e) = do
	t2 <- typeCheckExp e
	if t == t2 then return t else fail "Something went wrong ExpT is created with the wrong type"
typeCheckExp EDefault = return NoType
typeCheckExp (EPost i op) = do
	TInt <- typeCheckVar i
	return TInt
typeCheckExp (ENeg e) = do
	TInt <- typeCheckExp e
	return TInt
typeCheckExp (ENot e) = do
	TBool <- typeCheckExp e
	return TBool

typeCheckVar :: (MonadState Types m) => Ident -> m Type 
typeCheckVar i = do
	ms <- get
	findVariable i ms

findVariable :: (MonadState Types m) => Ident -> Types -> m Type
findVariable i [] = fail $ "Variable "++show i++" not found in any scope."
findVariable i (m:ms) = if member i m then lookup i m else findVariable i ms

typeCheckStm :: (MonadState Types m) => Stm -> m Type 
typeCheckStm SNoop = return NoType
typeCheckStm (SExp e) = do 
	typeCheckExp e
	return NoType
typeCheckStm (SBlock ss) = do 
	modify (empty:)
	mapM typeCheckStm ss
	modify tail
	return NoType
typeCheckStm (SIf e s s') = do
	TBool <- typeCheckExp e
	NoType <- typeCheckStm s
	NoType <- typeCheckStm s
	return NoType
typeCheckStm (SWhile e s) = do
	TBool <- typeCheckExp e
	NoType <- typeCheckStm s
	return NoType
typeCheckStm (SDecl t i e) = do
	t2 <- typeCheckExp e
	if t == t2 || t2 == NoType then do
		(m:ms) <- get
		case insertLookupWithKey (\k a1 a2 -> a1) i t m of
			(Nothing,m') -> put (m':ms)
			_ -> fail $ "Duplicate variable declaration: "++show i
		return NoType
		else fail $ "Illegal to assign an expression of type "++show t2++" to variable "++show i++" of type "++show t
typeCheckStm (SPrint e) = do
	typeCheckExp e
	return NoType
